#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

from abc import ABC

import cv2
from flatdict import FlatDict
import numpy as np
from pandas import DataFrame
import re
import ros_numpy
import rosbag
import rospy
import sensor_msgs.msg
import yaml

from pydtk.models import BaseModel, register_model


@register_model(priority=1)
class GenericRosbagModel(BaseModel, ABC):
    """A generic model for a rosbag file."""

    _content_type = None  # allow any content-type
    _data_type = None  # allow any data-type
    _file_extensions = ['.bag']
    _contents = None
    _config = {
        'keys_to_exclude': [
            'header.seq',
            'header.stamp.secs',
            'header.stamp.nsecs',
            'header.frame_id'
        ]
    }

    def __init__(self, **kwargs):
        super(GenericRosbagModel, self).__init__(**kwargs)

    def _load(self, path, contents=None,
              start_timestamp=None, end_timestamp=None,
              target_frame_rate=None,
              **kwargs):
        """Load a rosbag file.

        Args:
            path (str): path to a rosbag file
            contents (str or dict): topic name to load
            start_timestamp (float): timestamp to start loading (not supported)
            end_timestamp (float): timestamp to end loading (not supported)

        """
        topic = None
        if isinstance(contents, str):
            topic = contents
        if isinstance(contents, dict):
            topic = next(iter(contents))
        if topic is None:
            raise ValueError('Topic name must be specified by the argument "contents"')
        start_time = rospy.Time(start_timestamp) if start_timestamp else start_timestamp
        end_time = rospy.Time(end_timestamp) if end_timestamp else end_timestamp

        timestamps, data = [], []
        with rosbag.Bag(path, 'r') as bag:
            if target_frame_rate:
                timestamps = self.load_timestamps(bag, topic, start_time, end_time)
                timestamps = self.downsample_timestamps(timestamps,
                                                        target_frame_rate=target_frame_rate)
                idx = 0
            for topic, msg, t in bag.read_messages(topics=[topic], start_time=start_time,
                                                   end_time=end_time):
                timestamp = self.msg_to_timestamp(msg, t).to_sec()
                if target_frame_rate:
                    if timestamp == timestamps[idx]:
                        data.append(self.msg_to_data(msg, config=self._config, **kwargs))
                        idx += 1
                        if idx == len(timestamps):
                            break
                    continue
                timestamps.append(timestamp)
                data.append(self.msg_to_data(msg, config=self._config, **kwargs))

        self.data = {'timestamps': timestamps, 'data': data}

    def _load_as_generator(self, path, contents=None,
                           start_timestamp=None, end_timestamp=None,
                           target_frame_rate=None,
                           **kwargs):
        """Load a rosbag file for each sample.

        Args:
            path (str): path to a rosbag file
            contents (str or dict): topic name to load
            start_timestamp (float): timestamp to start loading (not supported)
            end_timestamp (float): timestamp to end loading (not supported)

        """
        topic = None
        if isinstance(contents, str):
            topic = contents
        if isinstance(contents, dict):
            topic = next(iter(contents))
        if topic is None:
            raise ValueError('Topic name must be specified by the argument "contents"')
        start_time = rospy.Time(start_timestamp) if start_timestamp else start_timestamp
        end_time = rospy.Time(end_timestamp) if end_timestamp else end_timestamp

        timestamps = []
        with rosbag.Bag(path, 'r') as bag:
            if target_frame_rate:
                timestamps = self.load_timestamps(bag, topic, start_time, end_time)
                timestamps = self.downsample_timestamps(timestamps,
                                                        target_frame_rate=target_frame_rate)
                idx = 0
            for topic, msg, t in bag.read_messages(topics=[topic], start_time=start_time,
                                                   end_time=end_time):
                timestamp = self.msg_to_timestamp(msg, t).to_sec()
                if target_frame_rate:
                    if timestamp == timestamps[idx]:
                        yield {
                            'timestamps': [timestamp],
                            'data': [self.msg_to_data(msg, config=self._config, **kwargs)]
                        }
                        idx += 1
                        if idx == len(timestamps):
                            break
                    continue
                yield {
                    'timestamps': [timestamp],
                    'data': [self.msg_to_data(msg, config=self._config, **kwargs)]
                }

    def _save(self, path, contents=None, **kwargs):
        """Save ndarray data to a csv file.

        Args:
            path (str): path to the output csv file
            contents (str or dict): topic name

        """
        pass

    def to_dataframe(self):
        """Return data as a Pandas DataFrame.

        Returns:
            (DataFrame): data

        """
        df = DataFrame.from_dict(self.data['data'])
        return df

    def to_ndarray(self):
        """Return data as ndarray."""
        df = self.to_dataframe()
        return df.to_numpy()

    def load_timestamps(self, bag, topic, start_time, end_time):
        """Load only timestamps."""
        timestamps = []
        for topic, msg, t in bag.read_messages(topics=[topic], start_time=start_time,
                                               end_time=end_time):
            timestamps.append(self.msg_to_timestamp(msg, t).to_sec())
        return timestamps

    @property
    def timestamps(self):
        """Return timestamps as ndarray."""
        return np.array(self.data['timestamps'])

    @property
    def columns(self):
        """Return columns."""
        if self._columns is not None:
            return self._columns

        if self.data is not None:
            if len(self.data['data']) > 0:
                return list(self.data['data'][0].keys())

        return []

    @staticmethod
    def msg_to_data(msg, **kwargs):
        """Convert a message to data.

        Args:
            msg (a ROS message): message

        Returns:
            (object): data

        """
        keys_to_exclude = []
        if 'config' in kwargs.keys() and 'keys_to_exclude' in kwargs['config'].keys():
            keys_to_exclude = kwargs['config']['keys_to_exclude']
        return {k: v for k, v in dict(FlatDict(yaml.safe_load(str(msg)), delimiter='.')).items() if
                k not in keys_to_exclude}

    @staticmethod
    def msg_to_timestamp(msg, t):
        """Extract timestamp from a message.

        Args:
            msg (a ROS message): message
            t (rospy.Time): timestamp read from rosbag

        Returns:
            (rospy.Time): timestamp

        """
        if hasattr(msg, 'header'):
            if msg.header.stamp.is_zero():
                return t
            return msg.header.stamp
        else:
            return t

    @classmethod
    def generate_contents_meta(cls, path, content_key='content'):
        """Generate contents metadata.

        Args:
            path (str): File path
            content_key (str): Key of content

        Returns:
            (dict): contents metadata

        """
        # Load file
        contents = {}
        with rosbag.Bag(path, "r") as bag:
            # get topic names
            topic_info = bag.get_type_and_topic_info()
        topics = [topic for topic in topic_info[1].keys()]
        topics.sort()

        # Generate metadata
        for topic in topics:
            contents[topic] = {}
            contents[topic]["msg_type"] = topic_info[1][topic].msg_type
            contents[topic]["msg_md5sum"] = topic_info[0][topic_info[1][topic].msg_type]
            contents[topic]["count"] = topic_info[1][topic].message_count
            contents[topic]["frequency"] = topic_info[1][topic].frequency
            contents[topic]["tags"] = re.split('[_/-]', topic[1:])

        return contents

    @classmethod
    def generate_timestamp_meta(cls, path):
        """Generate contents metadata.

        Args:
            path (str): File path

        Returns:
            (list): [start_timestamp, end_timestamp]

        """
        with rosbag.Bag(path, "r") as bag:
            start_time = bag.get_start_time()
            end_time = bag.get_end_time()
        return [start_time, end_time]


@register_model(priority=2)
class SensorMsgsCompressedImageRosbagModel(GenericRosbagModel, ABC):
    """A model for a rosbag file containing sensor_msgs/Range."""

    _contents = {'.*': {'msg_type': 'sensor_msgs/CompressedImage'}}
    _columns = ['red', 'green', 'blue']

    @staticmethod
    def msg_to_data(msg, resize_rate=1.0, **kwargs):
        """Convert a message to data."""
        jpg = np.fromstring(msg.data, np.uint8)
        image = cv2.imdecode(jpg, cv2.IMREAD_COLOR)
        if resize_rate != 1.0:
            image = cv2.resize(image, dsize=None, fx=resize_rate,
                               fy=resize_rate, interpolation=cv2.INTER_LINEAR)
        image = image[:, :, ::-1]  # Convert BGR to RGB
        image = image.transpose((2, 0, 1))  # Reshape: [H, W, C] -> [C, H, W]
        return image

    def to_ndarray(self):
        """Return data as ndarray."""
        return np.array(self.data['data'])


@register_model(priority=2)
class SensorMsgsPointCloud2RosbagModel(GenericRosbagModel, ABC):
    """A model for a rosbag file containing sensor_msgs/PointCloud2."""

    _contents = {'.*': {'msg_type': 'sensor_msgs/PointCloud2'}}
    _config = {'fields': ('x', 'y', 'z')}

    def __init__(self, fields=('x', 'y', 'z'), **kwargs):
        super(SensorMsgsPointCloud2RosbagModel, self).__init__(**kwargs)
        self._config['fields'] = fields

    def msg_to_data(self, msg, **kwargs):
        """Convert a message to data."""
        if msg.__class__.__name__ == "_sensor_msgs__PointCloud2":
            msg.__class__ = sensor_msgs.msg._PointCloud2.PointCloud2
        points = ros_numpy.numpify(msg)[list(self._config['fields'])]
        pointcloud = np.array(points.tolist())
        if 'intensity' in self._config['fields']:
            pointcloud[:, self._config['fields'].index('intensity')] /= 255.0  # scale to [0, 1]
        return pointcloud

    def to_ndarray(self):
        """Return data as ndarray."""
        return np.array(self.data['data'], dtype="object")

    @property
    def columns(self):
        """Return columns."""
        return list(self._config['fields'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

"""Base Statistic Calculation module."""

import pandas as pd

from abc import ABCMeta

from pydtk.statistics import calculator


class BaseStatisticCalculation(metaclass=ABCMeta):
    """Base Statistic Calculation."""

    def __init__(self, target_span=60.0, sync_timestamps=False):
        """Initialize Base Statistics Calculation class.

        Args:
            target_span (float): interval of statistics calculation
            sync_timestamps (bool): if True, the output timestamps will
                                          start from 'timestamp // span * span'

        """
        self.target_span = target_span
        self.sync_timestamps = sync_timestamps

    def _get_calculator(self, dtype):
        """Get calculator by data type.

        Args:
            dtype (string): dtype of ndarray

        Returns:
            dtype_calculator (object): calculator class

        """
        kwargs = {'target_span': self.target_span, 'sync_timestamps': self.sync_timestamps}
        if "bool" in dtype:
            dtype_calculator = getattr(calculator, "BoolCalculator")(**kwargs)
        else:
            dtype_calculator = getattr(calculator, "FloatCalculator")(**kwargs)
        return dtype_calculator

    def calculate(self, timestamps, data, operation):
        """Divide and calculate statistics of divided data.

        Args:
            timestamps (ndarray): timestamps [sec]
            data (ndarray): input data
            operation (str): operation

        Returns:
            index_timestamps (ndarray): timestamps [sec]
            stat_data (ndarray): mean of input data

        """
        self.calculator = self._get_calculator(str(data.dtype))
        index_timestamps, stat_data = getattr(self.calculator, operation)(timestamps, data)
        return index_timestamps, stat_data

    def mean(self, timestamps, data):
        """Divide and return means of divided data."""
        return self.calculate(timestamps, data, "mean")

    def max(self, timestamps, data):
        """Divide and return maximum of divided data."""
        return self.calculate(timestamps, data, "max")

    def min(self, timestamps, data):
        """Divide and return minimum of divided data."""
        return self.calculate(timestamps, data, "min")

    def count(self, timestamps, data):
        """Divide and return count of True in divided data."""
        return self.calculate(timestamps, data, "count")

    def statistic_tables(self, timestamps, data, columns):
        """Make statistic tables.

        Args:
            timestamps (ndarray): timestamps [sec]
            data (ndarray): input data
            columns (str): columns of table

        Returns:
            df_dict (dict): dict includes statistic DataFrames

        """
        self.calculator = self._get_calculator(str(data.dtype))
        for i, operation in enumerate(self.calculator.operations):
            index_timestamps, stat_data = self.calculate(timestamps, data, operation)
            time_df = pd.DataFrame(data=index_timestamps, columns=["timestamp"])
            if i == 0:
                stat_df = time_df
            if stat_data.ndim == 1:
                stat_data = stat_data.reshape(-1, 1)
            ope_columns = [column + "/" + operation for column in columns]
            _stat_df = pd.DataFrame(data=stat_data, columns=ope_columns)
            _stat_df = pd.concat([time_df, _stat_df], axis=1)
            stat_df = pd.merge(stat_df, _stat_df, on="timestamp")
        return stat_df



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright Toolkit Authors

from abc import ABC

import bitstring

from pydtk.models import register_model
from pydtk.models.rosbag import GenericRosbagModel as _GenericRosbagModel
from pydtk.utils.can_decoder import CANDecoder


@register_model(priority=2)
class AutowareCanMsgsCANPacketRosbagModel(_GenericRosbagModel, ABC):
    """A model for a rosbag file containing autoware_can_msgs/CANPacket."""

    _contents = {'.*': {'msg_type': 'autoware_can_msgs/CANPacket'}}
    _config = {'path_to_assign_list': None}
    can_decoder = None

    def __init__(self, path_to_assign_list=None, **kwargs):
        super(AutowareCanMsgsCANPacketRosbagModel, self).__init__(**kwargs)
        self._config['path_to_assign_list'] = path_to_assign_list
        if path_to_assign_list is not None:
            self.can_decoder = CANDecoder(path_to_assign_list)

    def configure(self, path_to_assign_list=None, **kwargs):
        """Configure the model.

        Args:
            path_to_assign_list (str): Path to bit-assign list (csv)

        """
        if path_to_assign_list is not None:
            self._config['path_to_assign_list'] = path_to_assign_list
            self.can_decoder = CANDecoder(path_to_assign_list)

    def msg_to_data(self, msg, **kwargs):
        """Convert a message to data."""
        if self.can_decoder is not None:
            can_id = bitstring.ConstBitArray(uint=int(msg.id), length=32).hex.lstrip("0").lower()
            can_data = bitstring.ConstBitArray(bytes=msg.dat).hex

            data = []
            for bit_assign in self.can_decoder.bit_assign_info.signal_list.values():
                if can_id.replace(" ", "") == bit_assign.can_id:
                    data.append(self.can_decoder.unpack_data(can_data, bit_assign))
                else:
                    data.append(None)

            return data
        else:
            return [msg.count, msg.id, msg.len, msg.dat, msg.flag, msg.time]

    @property
    def columns(self):
        """Return columns."""
        if self.can_decoder is not None:
            return [bit_assign.data_label
                    for bit_assign in self.can_decoder.bit_assign_info.signal_list.values()]
        else:
            return ['count', 'id', 'len', 'dat', 'flag', 'time']



from functools import wraps
from .aggregation import AggregationGroupParser, AggregationParser
from .matching import (SchemaFreeParser, SchemaAwareParser, ParseError,
                       StringField, IntField, BoolField, IdField,
                       ListField, DictField, DateTimeField)

def find(expression, schema=None):
    '''
    Gets an <expression> and optional <schema>.
    <expression> should be a string of python code.
    <schema> should be a dictionary mapping field names to types.
    '''
    parser = SchemaFreeParser() if schema is None else SchemaAwareParser(schema)
    return parser.parse(expression)

class pipe_element(list):
    def __or__(self, other):
        return pipe_element(self + other)

def pipe(func):
    @wraps(func)
    def decorated(*a, **k):
        return pipe_element([func(*a, **k)])
    return decorated

def _parse_value(parser, value):
    if isinstance(value, str) or isinstance(value, unicode):
        return parser.parse(value)
    elif isinstance(value, int):
        return value
    else:
        raise ValueError('Unexpected type: {}'.format(value.__class__))
def _parse_dict(parser, dct):
    return dict([(k, _parse_value(parser, v)) for k, v in dct.items()])

@pipe
def group(_id, **kwargs):
    group = _parse_dict(parser=AggregationGroupParser(), dct=kwargs)
    if isinstance(_id, pipe_element):
        _id = _id[0]['$project']
    else:
        _id = AggregationParser().parse(_id)
    group['_id'] = _id
    return {'$group': group}

@pipe
def project(**kwargs):
    return {'$project': _parse_dict(parser=AggregationParser(), dct=kwargs)}

@pipe
def match(expression, schema=None):
    return {'$match': find(expression, schema)}

@pipe
def limit(number):
    if not isinstance(number, int):
        raise ValueError("aggregation 'limit' must be a number")
    return {'$limit': number}

@pipe
def skip(number):
    if not isinstance(number, int):
        raise ValueError("aggregation 'skip' must be a number")
    return {'$skip': number}

@pipe
def unwind(list_name):
    if not isinstance(list_name, str):
        raise ValueError("aggregation 'unwind' must be a str")
    return {'$unwind': '$' + list_name}

@pipe
def sort(fields):
    '''
    Gets a list of <fields> to sort by.
    Also supports getting a single string for sorting by one field.
    Reverse sort is supported by appending '-' to the field name.
    Example: sort(['age', '-height']) will sort by ascending age and descending height.
    '''
    from pymongo import ASCENDING, DESCENDING
    from bson import SON

    if isinstance(fields, str):
        fields = [fields]
    if not hasattr(fields, '__iter__'):
        raise ValueError("expected a list of strings or a string. not a {}".format(type(fields)))
    
    sort = []
    for field in fields:
        if field.startswith('-'):
            field = field[1:]
            sort.append((field, DESCENDING))
            continue
        elif field.startswith('+'):
            field = field[1:]
        sort.append((field, ASCENDING))
    return {'$sort': SON(sort)}



#!/usr/bin/env python3

# Copyright Toolkit Authors (Yusuke Adachi)

import argparse
import json
import logging
import pandas as pd
from pathlib import Path
import pickle
import time

from pydtk.utils import utils


def _make_record_id_list(file_df, content_df):
    """Show record id list.

    Returns:
        (list): List of record id.

    """
    record_id_list = file_df["record_id"].unique()
    duration_list, tags_list = [], []
    start_timestamp_list, end_timestamp_list = [], []
    for record_id in record_id_list:
        logging.debug("Making record id indexing: %s" % record_id)
        mask = file_df["record_id"].apply(lambda x: record_id in x)
        record_id_df = file_df[mask]
        start_timestamp, end_timestamp = utils.take_time_and(record_id_df)
        duration = max(0.0, end_timestamp - start_timestamp)
        tags = []
        for tag_list in content_df["tag"]:
            for tag in tag_list:
                if tag not in tags:
                    tags += [tag]

        duration_list += [duration]
        start_timestamp_list += [start_timestamp]
        end_timestamp_list += [end_timestamp]
        tags_list += [tags]

    record_id_df = pd.DataFrame(
        data={'record_id': record_id_list,
              'duration': duration_list,
              'start_timestamp': start_timestamp_list,
              'end_timestamp': end_timestamp_list,
              'tags': tags_list
              },
        columns=['record_id', 'duration', 'start_timestamp', 'end_timestamp', 'tags']
    )
    return record_id_df


def find_json(db_dir):
    """Search with tags.

    Args:
        db_dir (str): Directory path which has database files.
    Returns:
        (list): List of json file path.

    """
    p = Path(db_dir)
    json_list = list(p.glob("**/*.json"))
    json_list.sort()
    return json_list


def _filter_json_list(json_list, start, end):
    """Now only support Driving Behavior Databse(DBDB)."""
    json_list_filtered = []
    for json_file in json_list:
        hdd_id = str(json_file).split("/")[-3].split("_")[0]
        if hdd_id.startswith('B') or hdd_id.startswith('W'):
            hdd_id = '400'
        if start <= int(hdd_id) <= end:
            json_list_filtered += [json_file]
    return json_list_filtered


def build_df_list(db_dir, pkl_file, start, end):
    """Search with tags.

    Args:
        db_dir (str): Directory path which has database files.
        pkl_file (str): Pickle file path to save.

    """
    logging.info("Making pickle: %s" % pkl_file)
    t0 = time.time()
    json_list = find_json(db_dir)
    if not (start == 0 and end == 999):
        json_list = _filter_json_list(json_list, start, end)

    t1 = time.time()
    logging.info("Find %d files.(%d secs)" % (len(json_list), t1 - t0))

    file_columns = ["path", "record_id", "type", "content-type", "start_timestamp", "end_timestamp"]
    # "description", "database_id"

    # Initialize list
    file_path_list, file_record_id_list = [], []
    file_type_list, file_content_type_list = [], []
    file_start_timestamp_list, file_end_timestamp_list = [], []

    content_record_id_list, content_path = [], []
    content_contents_list, content_tag_list = [], []
    content_msg_type_list = []

    for json_path in json_list:
        with open(json_path) as f:
            df = json.load(f)
        # Add file list
        file_path_list += [df["path"]]
        file_record_id_list += [df["record_id"]]
        file_type_list += [df["type"]]
        file_content_type_list += [df["content-type"]]
        file_start_timestamp_list += [df["start_timestamp"]]
        file_end_timestamp_list += [df["end_timestamp"]]

        # Add content list
        for content in df["contents"]:
            content_record_id_list += [df["record_id"]]
            content_path += [df["path"]]
            content_contents_list += [content]
            try:
                content_msg_type_list += [df["contents"][content]["msg_type"]]
            except KeyError:
                content_msg_type_list += ["None"]
            content_tag_list += [df["contents"][content]["tags"]]
    t2 = time.time()
    logging.info("Finish loading json.(%d secs)" % (t2 - t1))

    # 一気に列を結合したほうが早いらしい
    file_df = pd.DataFrame(
        data={'path': file_path_list,
              'record_id': file_record_id_list,
              'type': file_type_list,
              'content-type': file_content_type_list,
              'start_timestamp': file_start_timestamp_list,
              'end_timestamp': file_end_timestamp_list
              },
        columns=file_columns
    )

    content_df = pd.DataFrame(
        data={'record_id': content_record_id_list,
              'path': content_path,
              'content': content_contents_list,
              'msg_type': content_msg_type_list,
              'tag': content_tag_list
              },
        columns=["record_id", "path", "content", "msg_type", "tag"]
    )

    record_id_df = _make_record_id_list(file_df, content_df)
    t3 = time.time()
    logging.info("Finish making list of record id.(%d secs)" % (t3 - t2))

    with open(pkl_file, "wb") as f:
        pickle.dump({"file_df": file_df,
                     "content_df": content_df,
                     "record_id_df": record_id_df},
                    f, protocol=2)
    t4 = time.time()
    logging.info("Finish saving pickle file.(Total: %d secs)" % (t4 - t0))


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="This script is bababa")
    parser.add_argument('--db_dir', default="/data_pool_1/DrivingBehaviorDatabase/records",
                        help='Path to database directory')
    parser.add_argument('--pkl', default="pydtk_DBDB.pkl",
                        help='Pickle file path to save')
    parser.add_argument('--start', default=0, type=int,
                        help='Start HDD ID for index')
    parser.add_argument('--end', default=999, type=int,
                        help='End HDD ID for index')
    parser.add_argument('-v', '--verbose', action='store_true', help='Verbose mode')

    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    db_dir, pkl = args.db_dir, args.pkl
    start, end = args.start, args.end
    build_df_list(db_dir, pkl, start, end)
